"""Namespace for enum types etc."""

from enum import IntEnum


class FlowAlg(IntEnum):
    """Used for performance reasons"""

    DATA = 1
    EXEC = 2
    DATA_OPT = 3

    @staticmethod
    def str(mode):
        # not using __str__ here because FlowAlg only serves as an enum,
        # so there won't be any objects instantiated
        if mode == FlowAlg.DATA:
            return 'data'
        elif mode == FlowAlg.EXEC:
            return 'exec'
        elif mode == FlowAlg.DATA_OPT:
            return 'data opt'

        return None

    @staticmethod
    def from_str(mode):
        if mode == 'data':
            return FlowAlg.DATA
        elif mode == 'exec':
            return FlowAlg.EXEC
        elif mode == 'data opt':
            return FlowAlg.DATA_OPT

        return None


class PortObjPos(IntEnum):
    """Used for performance reasons"""

    INPUT = 1
    OUTPUT = 2



from .dtypes import DType

# TODO: make this a dataclass
class NodePortBP:
    """
    The NodePortBP classes are only placeholders (BP = BluePrint) for the static init_input and
    init_outputs of custom Node classes.
    An instantiated Node's actual inputs and outputs will be of type NodeObjPort (NodeObjInput, NodeObjOutput).
    """

    def __init__(self, label: str = '', type_: str = 'data'):

        self.type_: str = type_
        self.label: str = label


class NodeInputBP(NodePortBP):
    def __init__(self, label: str = '', type_: str = 'data', dtype: DType = None, add_data={}):
        super().__init__(label, type_)

        self.dtype = dtype
        self.add_data = add_data


class NodeOutputBP(NodePortBP):
    def __init__(self, label: str = '', type_: str = 'data'):
        super().__init__(label, type_)



def complete_data(data: dict) -> dict:
    """Default implementation for completing data with frontend properties.
    When running with a frontend that needs to store additional information about
    components (e.g. current position and color of a node), this frontend then
    overrode this exact function to add whatever is necessary for adding this data."""
    return data


class Event:
    def __init__(self, *args):
        self.args = args
        self._slots = []

    def connect(self, callback):
        self._slots.append(callback)

    def disconnect(self, callback):
        self._slots.remove(callback)

    def emit(self, *args):
        for cb in self._slots:
            cb(*args)


class Base:
    """
    Base class for all abstract components. Provides functionality for ID counting.
    Assigns a global ID to every object and provides an optional custom ID counter for additional custom counting.
    """

    class IDCtr:
        def __init__(self):
            self.ctr = -1

        def count(self):
            """increases the counter and returns the new count. first time is 0"""
            self.ctr += 1
            return self.ctr

        def set_count(self, cnt):
            if cnt < self.ctr:
                raise Exception("Decreasing ID counters is illegal")
            else:
                self.ctr = cnt

    # all abstract components have a global ID
    global_id_ctr = IDCtr()

    # optional custom ID counter
    id_ctr = None
    # notice that the attribute is static, but a subclass changing it will not change it for Base
    # and hence not for other Base subclasses, only for itself

    # events
    events = {}
    # format: {event_name : tuple_of_arguments}
    # the arguments tuple only serves documentation purposes

    def __init__(self):
        self.GLOBAL_ID = self.global_id_ctr.count()

        # ignore custom ID if it has already been set
        if self.id_ctr is not None and not (hasattr(self, 'ID') and self.ID is not None):
            self.ID = self.id_ctr.count()

        self._slots = {
            ev: []
            for ev in self.events
        }

    # CUSTOM DATA ------------------------------------

    # this can be set to another function by the frontend to implement adding frontend information to the data dict
    complete_data_function = lambda data: data

    def data(self) -> dict:
        """converts the object to a JSON compatible dict for serialization"""
        return None

    def complete_data(self, data: dict) -> data:
        return Base.complete_data_function(data)

    # EVENTS ------------------------------------

    def on(self, ev: Event, callback):
        # self._slots[ev].append(callback)
        ev.connect(callback)

    def off(self, ev: Event, callback):
        # self._slots[ev].remove(callback)
        ev.disconnect(callback)

    # def _emit(self, ev: str, *args, **kwargs):
    #     for s in self._slots[ev]:
    #         s(args, kwargs)



import logging
import traceback
from typing import List, Dict

from . import FlowAlg
from .Base import Base

from .NodePort import NodeInput, NodeOutput
from .NodePortBP import NodeInputBP, NodeOutputBP
from .dtypes import DType
from .InfoMsgs import InfoMsgs
from .logging import Logger
from .utils import serialize, deserialize


class Node(Base):
    """
    Base class for all node blueprints. Such a blueprint is made by subclassing this class and registering that subclass
    in the session. Actual node objects are instances of it. The node's static properties are static attributes,
    which works really well in Python.

    Attributes:
        static title: the node's title
        static type_: conditional type field
        static tags: a list of tag strings, usually for searching
        static version: version tag, use it!
        static init_inputs: initial inputs list. the type BP in NodeInputBP stands for 'blueprint', as they only
            serve as containers, the actual input objects will be created later
        static init_outputs: initial outputs list, see init_inputs
        static identifier: unique node identifier string. if not given, the session will set it to the class name
            on register
        static identifier_comp: a list of compatible identifiers, useful if you change the class name (and hence
            the identifier) to provide backward compatibility to older projects
    """

    title = ''
    type_ = ''
    tags: List[str] = []
    version: str = None  # None means `undefined` and should be avoided
    visible: bool = True  # useful field for frontends to indicate invisible nodes which cannot be manually placed
    
    init_inputs: List[NodeInputBP] = []
    init_outputs: List[NodeOutputBP] = []

    identifier: str = None  # set by Session if None
    identifier_comp: List[str] = []  # identifier (backwards) compatibility, useful when node class name changes
    identifier_prefix: str = None  # becomes part of identifier if set, often useful

    """
    INITIALIZATION -----------------------------------------------------------------------------------------------------
    """

    @classmethod
    def build_identifier(cls):
        """
        sets the identifier to class name if it's not set, and adds the identifier prefix
        """

        prefix = ''
        if cls.identifier_prefix is not None:
            prefix = cls.identifier_prefix + '.'

        if cls.identifier is None:
            cls.identifier = cls.__name__

        cls.identifier = prefix + cls.identifier

        # notice that we do not touch the identifier compatibility fields

    def __init__(self, params):
        Base.__init__(self)

        self.flow, self.session, self.init_data = params
        self.script = self.flow.script
        self.inputs: List[NodeInput] = []
        self.outputs: List[NodeOutput] = []
        self.loggers = []

        self.initialized = False

        self.block_init_updates = False
        self.block_updates = False

    def initialize(self):
        """
        Loads all default properties from initial data if it was provided,
        sets up inputs and outputs, enables the logs, and loads user_data (doesn't crash on exception here
        as this is not uncommon when developing nodes).
        """

        # enable logs
        self.enable_loggers()

        if self.init_data:  # load from data
            # setup ports
            self.setup_ports(self.init_data['inputs'], self.init_data['outputs'])

            # set state
            if 'additional data' in self.init_data:
                add_data = self.init_data['additional data']
            else:   # backwards compatibility
                add_data = self.init_data
            self.load_additional_data(add_data)

            try:
                if 'version' in self.init_data:
                    version = self.init_data['version']
                else:  # backwards compatibility
                    version = None

                self.set_state(deserialize(self.init_data['state data']), version)

            except Exception as e:
                InfoMsgs.write_err(
                    'Exception while setting data in', self.title, 'node:', e, ' (was this intended?)')

        else:   # default setup

            # setup ports
            self.setup_ports()

        self.initialized = True

    def setup_ports(self, inputs_data=None, outputs_data=None):

        if not inputs_data and not outputs_data:
            # generate initial ports

            for i in range(len(self.init_inputs)):
                inp = self.init_inputs[i]

                if inp.dtype:
                    self.create_input_dt(dtype=inp.dtype, label=inp.label, add_data=inp.add_data)

                else:
                    self.create_input(inp.label, inp.type_, add_data=self.init_inputs[i].add_data)

            for o in range(len(self.init_outputs)):
                out = self.init_outputs[o]
                self.create_output(out.label, out.type_)

        else:
            # load from data
            # initial ports specifications are irrelevant then

            for inp in inputs_data:
                if 'dtype' in inp:
                    self.create_input_dt(dtype=DType.from_str(inp['dtype'])(
                        _load_state=deserialize(inp['dtype state'])), label=inp['label'], add_data=inp)
                else:
                    self.create_input(label=inp['label'], type_=inp['type'], add_data=inp)

                if 'val' in inp:
                    # this means the input is 'data' and did not have any connections,
                    # so we saved its value which was probably represented by some widget
                    # in the front end which has probably overridden the Node.input() method
                    self.inputs[-1].val = deserialize(inp['val'])

            for out in outputs_data:
                self.create_output(out['label'], out['type'])

    def after_placement(self):
        """Called from Flow when the nodes gets added"""

        self.enable_loggers()
        self.place_event()

    def prepare_removal(self):
        """Called from Flow when the node gets removed"""

        self.remove_event()
        self.disable_loggers()

    """
    ALGORITHM ----------------------------------------------------------------------------------------------------------
    """

    # notice that all the below methods check whether the flow currently 'runs with an executor', which means
    # the flow is running in a special execution mode, in which case all the algorithm-related methods below are
    # handled by the according executor

    def update(self, inp=-1):  # , output_called=-1):
        """'Activates' the node, causing an update_event(); prints an exception if something crashed, but prevents
        the application from crashing in such a case"""

        if self.block_updates:
            InfoMsgs.write('update blocked in', self.title, 'node')
            return

        InfoMsgs.write('update in', self.title, 'node on input', inp)

        # invoke update_event
        if self.flow.running_with_executor:
            self.flow.executor.update_node(self, inp)
        else:
            try:
                self.update_event(inp)
            except Exception as e:
                InfoMsgs.write_err('EXCEPTION in', self.title, '\n', traceback.format_exc())

    def input(self, index: int):
        """Returns the value of a data input.
        If the input is connected, the value of the connected output is used:
        If not, the value of the widget is used."""

        InfoMsgs.write('input called in', self.title, 'Node:', index)

        if self.flow.running_with_executor:
            return self.flow.executor.input(self, index)
        else:
            return self.inputs[index].get_val()

    def exec_output(self, index: int):
        """Executes an exec output, causing activation of all connections"""
        if self.flow.running_with_executor:
            self.flow.executor.exec_output(self, index)
        else:
            self.outputs[index].exec()

    def set_output_val(self, index, val):
        """Sets the value of a data output causing activation of all connections in data mode"""
        if self.flow.running_with_executor:
            self.flow.executor.set_output_val(self, index, val)
        else:
            self.outputs[index].set_val(val)

    """
    EVENTS -------------------------------------------------------------------------------------------------------------
    """

    # these methods get implemented by node implementations

    def update_event(self, inp=-1):
        """
        Gets called when an input received a signal or some node requested data of an output in exec mode
        """

        pass

    def place_event(self):
        """
        place_event() is called once the node object has been fully initialized and placed in the flow.
        When loading content, place_event() is executed *before* the connections are built,
        which is important for nodes that need to update once and, during this process, set output data values,
        to prevent later connected (potentially sequential) nodes from receiving false updates because of that.
        Notice that this method gets executed *every time* the node is added to the flow, which can happen
        multiple times for the same object, for example due to undo/redo operations.
        Also note that GUI content is usually not accessible yet from here, for that use view_place_event().
        """

        pass

    def view_place_event(self):
        """
        Called once all GUI for the node has been created by the frontend, if one exists.
        Any initial communication to widgets is supposed to happen here, and this method is not called
        when running without gui.
        """

        pass

    def remove_event(self):
        """
        Called when the node is removed from the flow; useful for stopping threads and timers etc.
        """

        pass

    def additional_data(self) -> dict:
        """
        Additional_data()/load_additional_data() is almost equivalent to get_state()/set_state(),
        but it's often useful for frontends to have their own,
        get_state()/set_state() then stays clean for all specific node subclasses
        """

        return {}

    def load_additional_data(self, data: dict):
        """
        For loading the data returned by additional_data()
        """
        pass

    def get_state(self) -> dict:
        """
        Used to store node-specific custom data that needs to be reloaded when loading a project or pasting copied
        components. All values will be serialized by pickle and base64. The corresponding method for the opposite
        operation is set_state().
        """
        return {}

    def set_state(self, data: dict, version):
        """
        Used for reloading node-specific custom data which has been previously returned by get_state()
        """
        pass

    """
    API ----------------------------------------------------------------------------------------------------------------
    """

    #   LOGGING

    def new_logger(self, title) -> Logger:
        """Requesting a new custom Log"""

        logger = self.script.logs_manager.new_logger(title)
        self.loggers.append(logger)
        return logger

    def disable_loggers(self):
        """Disables custom logs"""

        for logger in self.loggers:
            logger.disable()
            # logger.disabled = True

    def enable_loggers(self):
        """Enables custom logs"""

        for logger in self.loggers:
            logger.enable()
            # logger.enabled = True

    #   PORTS

    def create_input(self, label: str = '', type_: str = 'data', add_data={}, insert: int = None):
        """Creates and adds a new input at index pos"""
        # InfoMsgs.write('create_input called')

        inp = NodeInput(
            node=self,
            type_=type_,
            label_str=label,
            add_data=add_data,
        )

        if insert is not None:
            self.inputs.insert(insert, inp)
        else:
            self.inputs.append(inp)

    def create_input_dt(self, dtype: DType, label: str = '', add_data={}, insert: int = None):
        """Creates and adds a new data input with a DType"""
        # InfoMsgs.write('create_input called')

        inp = NodeInput(
            node=self,
            type_='data',
            label_str=label,
            dtype=dtype,
            add_data=add_data,
        )

        if insert is not None:
            self.inputs.insert(insert, inp)
        else:
            self.inputs.append(inp)

    def rename_input(self, index: int, label: str):
        self.inputs[index].label_str = label

    def delete_input(self, index: int):
        """Disconnects and removes input"""

        inp: NodeInput = self.inputs[index]

        # break all connections
        for c in inp.connections:
            self.flow.connect_nodes(c.out, inp)

        self.inputs.remove(inp)

    def create_output(self, label: str = '', type_: str = 'data', insert: int = None):
        """Creates and adds a new output"""

        out = NodeOutput(
              node=self,
              type_=type_,
              label_str=label
        )

        if insert is not None:
            self.outputs.insert(insert, out)
        else:
            self.outputs.append(out)

    def rename_output(self, index: int, label: str):
        self.outputs[index].label_str = label

    def delete_output(self, index: int):
        """Disconnects and removes output"""

        out: NodeOutput = self.outputs[index]

        # break all connections
        for c in out.connections:
            self.flow.connect_nodes(out, c.inp)

        self.outputs.remove(out)

    #   VARIABLES

    def get_vars_manager(self):
        """Returns a ref to the script's variables manager"""

        return self.script.vars_manager

    def get_var_val(self, name: str):
        """Gets the value of a script variable"""

        return self.get_vars_manager().get_var_val(name)

    def set_var_val(self, name: str, val):
        """Sets the value of a script variable"""

        return self.get_vars_manager().set_var(name, val)

    def register_var_receiver(self, name: str, method):
        """Registers the node with given method as vars receiver in the script's variables manager to catch
        value changes of any variable with the given name"""

        self.get_vars_manager().register_receiver(self, name, method)

    def unregister_var_receiver(self, name: str, method):
        """Unregisters previously registered node as receiver for value changes of script variables with given name"""

        self.get_vars_manager().unregister_receiver(self, name, method)

    """
    --------------------------------------------------------------------------------------------------------------------
    """

    def is_active(self):
        for i in self.inputs:
            if i.type_ == 'exec':
                return True
        for o in self.outputs:
            if o.type_ == 'exec':
                return True
        return False

    def flow_in_data_opt_mode(self):
        return self.flow.alg_mode == FlowAlg.DATA_OPT

    def data(self) -> dict:
        """
        Returns all metadata of the node in JSON-compatible dict.
        Used to rebuild the Flow when loading a project or pasting components.
        """

        return {
            'identifier': self.identifier,
            'version': self.version,

            'state data': serialize(self.get_state()),
            'additional data': self.additional_data(),

            'inputs': [i.data() for i in self.inputs],
            'outputs': [o.data() for o in self.outputs],

            'GID': self.GLOBAL_ID,
        }



from .Base import Base
from .logging import LogsManager
from .script_variables import VarsManager
from .Flow import Flow


class Script(Base):
    """A Script consists of a Flow, a VarsManager and a Logger object."""

    def __init__(self, session, title: str = None, load_data: dict = None, create_default_logs=True):
        Base.__init__(self)

        self.session = session
        self.logs_manager = None
        self._create_default_logs = create_default_logs
        self.vars_manager = None
        self.title = title
        self.flow = None

        if title is None and load_data is not None:
            self.title = load_data['title'] \
                if 'title' in load_data else load_data['name']  # backwards compatibility

        self.init_data = load_data
        self.init_flow_data = None
        self.init_vars_manager_data = None

        # loading from saved data
        if load_data:
            self.init_flow_data = load_data['flow'] if load_data else None
            self.init_vars_manager_data = load_data['variables']

        # logging
        self.logs_manager = LogsManager(self, self._create_default_logs)

        # vars manager
        self.vars_manager = VarsManager(self, self.init_vars_manager_data)

        # flow
        self.flow = Flow(self.session, self)


    def load_flow(self):
        if self.init_flow_data:
            self.flow.load(self.init_flow_data)


    def data(self) -> dict:
        return {
            'title': self.title,
            'variables': self.vars_manager.data(),
            'flow': self.flow.data(),
            'GID': self.GLOBAL_ID,
        }



from .Base import Base, Event


from .Script import Script
from .InfoMsgs import InfoMsgs

from typing import List, Dict


class Session(Base):
    """
    The Session is the top level interface to your project. It mainly manages Scripts and registered nodes, and
    provides methods for serialization and deserialization of the project.
    """

    def __init__(
            self,
            gui: bool = False,
    ):
        Base.__init__(self)

        # events
        self.new_script_created = Event(Script)
        self.script_renamed = Event(Script)
        self.script_deleted = Event(Script)

        # ATTRIBUTES
        self.scripts: [Script] = []
        self.nodes = []  # list of node CLASSES
        self.invisible_nodes = []
        self.gui: bool = gui
        self.init_data = None


    def register_nodes(self, node_classes: List):
        """Registers a list of Nodes which then become available in the flows"""

        for n in node_classes:
            self.register_node(n)


    def register_node(self, node_class):
        """Registers a single Node which then becomes available in the flows"""

        # build node class identifier
        node_class.build_identifier()

        self.nodes.append(node_class)


    def unregister_node(self, node_class):
        """Unregisters a Node which will then be removed from the available list.
        Existing instances won't be affected."""

        self.nodes.remove(node_class)


    def all_node_objects(self) -> List:
        """Returns a list of all Node objects instantiated in any flow"""

        nodes = []
        for s in self.scripts:
            for n in s.flow.nodes:
                nodes.append(n)
        return nodes


    def create_script(self, title: str = None, create_default_logs=True,
                      data: Dict = None) -> Script:
        """Creates and returns a new script.
        If data is provided the title parameter will be ignored."""

        script = Script(
            session=self, title=title, create_default_logs=create_default_logs,
            load_data=data
        )

        self.scripts.append(script)
        script.load_flow()

        self.new_script_created.emit(script)

        return script


    def rename_script(self, script: Script, title: str) -> bool:
        """Renames an existing script and returns success boolean"""

        success = False

        if self.script_title_valid(title):
            script.title = title
            success = True

        self.script_renamed.emit(script)

        return success


    def script_title_valid(self, title: str) -> bool:
        """Checks whether a considered title for a new script is valid (unique) or not"""

        if len(title) == 0:
            return False
        for s in self.scripts:
            if s.title == title:
                return False

        return True


    def delete_script(self, script: Script):
        """Removes an existing script."""

        self.scripts.remove(script)

        self.script_deleted.emit(script)


    def info_messenger(self):
        """Returns a reference to InfoMsgs to print info data"""

        return InfoMsgs


    def load(self, project: Dict) -> List[Script]:
        """Loads a project and raises an exception if required nodes are missing"""

        # TODO: perform validity checks

        self.init_data = project

        new_scripts = []
        for sc in project['scripts']:
            new_scripts.append(self.create_script(data=sc))

        return new_scripts

    def serialize(self):
        """Returns the project as JSON compatible dict to be saved and loaded again using load()"""

        return self.complete_data(self.data())


    def data(self) -> dict:
        return {
            'scripts': [
                s.data() for s in self.scripts
            ],
        }



"""
While standard flow execution (data-flow or exec-flow) is implemented inside the classes (like Node),
this module provides special FlowExecutors which take over most of the work and implement more
sophisticated and optimized flow execution.
"""


class FlowExecutor:
    """
    Base class for special flow execution algorithms.
    """

    def __init__(self, flow):
        self.flow = flow

    # Node.update() =>
    def update_node(self, node, inp):
        pass

    # Node.input() =>
    def input(self, node, index):
        pass

    # Node.set_output_val() =>
    def set_output_val(self, node, index, val):
        pass

    # Node.exec_output() =>
    def exec_output(self, node, index):
        pass


class DataFlowOptimized(FlowExecutor):
    """
    A special flow executor which implements some node functions to optimise flow execution.
    Whenever a new execution is invoked somewhere (some node or output is updated), it
    analyses the graph's connected component (of successors) where the execution was invoked
    and creates a few data structures to reverse engineer how many input
    updates every node possibly receives in this execution. A node's outputs are
    propagated once no input can still receive new data from a predecessor node.
    Therefore, while a node gets updated every time an input receives some data,
    every OUTPUT is only updated ONCE.
    This implies that every connection is activated at most once in an execution.
    This can result in asymptotic speedup in large data flows compared to normal data flow
    execution where any two executed branches which merge again in the future result in two
    complete executions of everything that comes after the merge, which quickly produces
    exponential performance issues.
    """

    def __init__(self, flow):
        super().__init__(flow)

        self.output_updated = {}
        self.waiting_count = {}
        self.node_waiting = {}
        self.num_conns_from_predecessors = None
        self.last_execution_root = None     # for reuse when a same execution is invoked many times consecutively
        self.execution_root = None          # can be Node or NodeOutput
        self.execution_root_node = None     # the updated Node or the updated NodeOutput's Node
        self.flow_changed = True

    # NODE FUNCTIONS

    # Node.update() =>
    def update_node(self, node, inp=-1):
        if self.execution_root_node is None:  # execution starter!
            self.start_execution(root_node=node)
            self.invoke_node_update_event(node, inp)
            self.propagate_outputs(node)
            self.stop_execution()
        else:
            self.invoke_node_update_event(node, inp)

    # Node.input() =>
    def input(self, node, index):
        return node.inputs[index].get_val()

    # Node.set_output_val() =>
    def set_output_val(self, node, index, val):
        out = node.outputs[index]

        if self.execution_root_node is None:  # execution starter!
            self.start_execution(root_output=out)

            out.val = val
            self.output_updated[out] = True
            self.propagate_output(out)

            self.stop_execution()

        else:

            if not self.node_waiting[out.node]:
                # the output's node might not be part of the analyzed graph!
                # in this case we immediately push the value
                # there are other possible solutions to this, including running
                # a new execution analysis of this graph here

                out.set_val(val)

            else:
                out.val = val
                self.output_updated[out] = True


    # Node.exec_output() =>
    def exec_output(self, node, index):
        out = node.outputs[index]

        if self.execution_root_node is None:  # execution starter!
            self.start_execution(root_output=out)

            self.output_updated[out] = True
            self.propagate_output(out)

            self.stop_execution()

        else:
            self.output_updated[out] = True

    # ----------------------------------------------------------

    def start_execution(self, root_node=None, root_output=None):

        # reset cached output values
        self.output_updated = {}
        for n in self.flow.nodes:
            for out in n.outputs:
                self.output_updated[out] = False

        if root_node is not None:
            self.execution_root = root_node
            self.execution_root_node = root_node
            self.waiting_count = self.generate_waiting_count(root_node=root_node)

        elif root_output is not None:
            self.execution_root = root_output
            self.execution_root_node = root_output.node
            self.waiting_count = self.generate_waiting_count(root_output=root_output)

    def stop_execution(self):
        self.execution_root_node = None
        self.last_execution_root = self.execution_root
        self.execution_root = None

    def generate_waiting_count(self, root_node=None, root_output=None):
        if not self.flow_changed and self.execution_root is self.last_execution_root:
            return self.num_conns_from_predecessors.copy()
        self.flow_changed = False

        nodes = self.flow.nodes
        node_successors = self.flow.node_successors

        # DP TABLE
        self.num_conns_from_predecessors = {
            n: 0
            for n in nodes
        }

        successors = set()
        visited = {
            n: False
            for n in nodes
        }

        # BC
        if root_node is not None:
            successors.add(root_node)
        elif root_output is not None:
            for c in root_output.connections:
                connected_node = c.inp.node
                self.num_conns_from_predecessors[connected_node] += 1
                successors.add(connected_node)

        # ITERATION
        while len(successors) > 0:
            n = successors.pop()
            if visited[n]:
                continue

            for s in node_successors[n]:
                self.num_conns_from_predecessors[s] += 1
                successors.add(s)
            visited[n] = True

        self.node_waiting = visited

        return self.num_conns_from_predecessors.copy()

    def invoke_node_update_event(self, node, inp):
        try:
            node.update_event(inp)
        except Exception:
            pass

    def decrease_wait(self, node):
        """decreases the wait count of the node;
        if the count reaches zero, which means there is no other input waiting for data,
        the output values get propagated"""

        self.waiting_count[node] -= 1
        if self.waiting_count[node] == 0:
            self.propagate_outputs(node)

    def propagate_outputs(self, node):
        """propagates all outputs of node"""

        for out in node.outputs:
            self.propagate_output(out)

    def propagate_output(self, out):
        """pushes an output's value to successors if it has been changed in the execution"""

        if self.output_updated[out]:

            if out.type_ == 'data':         # data output updated
                for c in out.connections:
                    c.activate(out.val)

            else:                           # exec output executed
                for c in out.connections:
                    c.activate()

        # decrease wait count of successors
        for c in out.connections:
            self.decrease_wait(c.inp.node)



from .Base import Base, Event

from .InfoMsgs import InfoMsgs


class Connection(Base):
    """
    The base class for both types of connections. All data is transmitted through a connection from an output
    port to some connected input port.
    """

    def __init__(self, params):
        Base.__init__(self)

        self.activated = Event(object)

        self.out, self.inp, self.flow = params

    def activate(self, data=None):
        """Causes forward propagation of information"""

        self.activated.emit(data)


class ExecConnection(Connection):

    def activate(self, data=None):
        """Causes an update in the input port"""
        InfoMsgs.write('exec connection activated')
        super().activate()

        self.inp.update()


class DataConnection(Connection):

    def __init__(self, params):
        super().__init__(params)

        self.data = None

    def get_val(self):
        """Gets the value of the output port -- only used in exec mode flows"""
        InfoMsgs.write('data connection getting value')

        # request data backwards
        self.data = self.out.get_val()

        return self.data

    def activate(self, data=None):
        """Passes data to the input port and causes update"""
        InfoMsgs.write('data connection activated')
        super().activate(data)

        # store data
        self.data = data

        # propagate data forward
        self.inp.update(data)



from logging import Logger as PyLogger
from ..Base import Base, Event


class Logger(Base, PyLogger):
    """
    A small wrapper template for the python loggers to add functionality on node events.
    Reimplemented as wrapper by the frontend with according implementations of the below methods.
    """

    def __init__(self, *args, **kwargs):
        Base.__init__(self)
        PyLogger.__init__(self, *args, **kwargs)

        # events
        self.sig_enabled = Event()
        self.sig_disabled = Event()

    def enable(self):
        self.sig_enabled.emit()

    def disable(self):
        self.sig_disabled.emit()



from .Base import Base, Event
from .Connection import Connection, DataConnection, ExecConnection
from .FlowExecutor import DataFlowOptimized, FlowExecutor
from .Node import Node
from .NodePort import NodePort
from .RC import FlowAlg, PortObjPos
from .utils import node_from_identifier
from typing import List, Dict, Optional


class Flow(Base):
    """
    Manages all abstract flow components (nodes, connections) and includes implementations for editing.
    """

    def __init__(self, session, script):
        Base.__init__(self)

        # events
        self.node_added = Event(Node)
        self.node_removed = Event(Node)
        self.connection_added = Event(Connection)
        self.connection_removed = Event(Connection)

        self.connection_request_valid = Event(bool)
        self.nodes_created_from_data = Event(list)
        self.connections_created_from_data = Event(list)

        self.algorithm_mode_changed = Event(str)

        # general attributes
        self.session = session
        self.script = script
        self.nodes: [Node] = []
        self.connections: [Connection] = []

        self.alg_mode = FlowAlg.DATA

        # special executors
        self.executor_data_opt = DataFlowOptimized(self)
        self.executor: FlowExecutor = None
        self.running_with_executor = False
        self._update_running_with_executor()
        #   additional data structures for executors
        self.node_successors = {}


    def load(self, data):
        """Loading a flow from data"""

        # algorithm mode
        mode = data['algorithm mode']
        if mode == 'data' or mode == 'data flow':
            self.set_algorithm_mode('data')
        elif mode == 'data opt':
            self.set_algorithm_mode('data opt')
        elif mode == 'exec' or mode == 'exec flow':
            self.set_algorithm_mode('exec')

        # build flow

        new_nodes = self.create_nodes_from_data(data['nodes'])

        #   the following connections should not cause updates in sequential nodes
        blocked_nodes = filter(lambda n: n.block_init_updates, new_nodes)
        for node in blocked_nodes:
            node.block_updates = True

        self.connect_nodes_from_data(new_nodes, data['connections'])

        for node in blocked_nodes:
            node.block_updates = False


    def create_nodes_from_data(self, nodes_data: List):
        """Creates Nodes from nodes_data, previously returned by data()"""

        nodes = []

        for n_c in nodes_data:

            # find class
            node_class = node_from_identifier(
                n_c['identifier'],
                self.session.nodes + self.session.invisible_nodes
            )

            node = self.create_node(node_class, n_c)
            nodes.append(node)

        self.nodes_created_from_data.emit(nodes)

        return nodes


    def create_node(self, node_class, data=None):
        """Creates, adds and returns a new node object"""

        node = node_class((self, self.session, data))
        # node.finish_initialization()
        # node.load_user_data()  # --> Node.set_state()
        node.initialize()
        self.add_node(node)
        return node


    def add_node(self, node: Node):
        """Stores a node object and causes the node's place_event()"""

        self.nodes.append(node)
        self.node_successors[node] = []
        node.after_placement()
        self.flow_changed()

        # self.emit_event('node added', (node,))    # ALPHA
        self.node_added.emit(node)


    def node_view_placed(self, node: Node):
        """Triggered after the node's GUI content has been fully initialized"""

        node.view_place_event()


    def remove_node(self, node: Node):
        """Removes a node from internal list without deleting it"""

        node.prepare_removal()
        self.nodes.remove(node)
        # del self.node_successors[node]
        self.flow_changed()

        # self.emit_event('node removed', (node,))    # ALPHA
        self.node_removed.emit(node)


    def connect_nodes_from_data(self, nodes: List[Node], data: List):
        connections = []

        for c in data:

            c_parent_node_index = c['parent node index']
            c_connected_node_index = c['connected node']
            c_output_port_index = c['output port index']
            c_connected_input_port_index = c['connected input port index']

            if c_connected_node_index is not None:  # which can be the case when pasting
                parent_node = nodes[c_parent_node_index]
                connected_node = nodes[c_connected_node_index]

                c = self.connect_nodes(parent_node.outputs[c_output_port_index],
                                       connected_node.inputs[c_connected_input_port_index])
                connections.append(c)

        self.connections_created_from_data.emit(connections)

        return connections


    def check_connection_validity(self, p1: NodePort, p2: NodePort) -> bool:
        """Checks whether a considered connect action is legal"""

        valid = True

        if p1.node == p2.node:
            valid = False

        if p1.io_pos == p2.io_pos or p1.type_ != p2.type_:
            valid = False

        self.connection_request_valid.emit(valid)

        return valid


    def connect_nodes(self, p1: NodePort, p2: NodePort) -> Connection:
        """Connects nodes or disconnects them if they are already connected"""

        if not self.check_connection_validity(p1, p2):
            return None

        out = p1
        inp = p2
        if out.io_pos == PortObjPos.INPUT:
            out, inp = inp, out

        for c in out.connections:
            if c.inp == inp:
                # DISCONNECT
                self.remove_connection(c)
                return None

        # c = self.session.CLASSES['data conn']((out, inp, self)) if out.type_ == 'data' else \
        #     self.session.CLASSES['exec conn']((out, inp, self))
        c = DataConnection((out, inp, self)) if out.type_ == 'data' else \
            ExecConnection((out, inp, self))

        self.add_connection(c)

        return c


    def add_connection(self, c: Connection):
        """Adds a connection object"""

        c.out.connections.append(c)
        c.inp.connections.append(c)
        c.out.connected()
        c.inp.connected()
        self.connections.append(c)

        self.node_successors[c.out.node].append(c.inp.node)
        self.flow_changed()

        # self.emit_event('connection added', (c,))    # ALPHA
        self.connection_added.emit(c)


    def remove_connection(self, c: Connection):
        """Removes a connection object without deleting it"""

        c.out.connections.remove(c)
        c.inp.connections.remove(c)
        c.out.disconnected()
        c.inp.disconnected()
        self.connections.remove(c)

        self.node_successors[c.out.node].remove(c.inp.node)
        self.flow_changed()

        # self.emit_event('connection removed', (c,))    # ALPHA
        self.connection_removed.emit(c)


    def algorithm_mode(self) -> str:
        """Returns the current algorithm mode of the flow as string"""

        return FlowAlg.str(self.alg_mode)


    def set_algorithm_mode(self, mode: str):
        """Sets the algorithm mode of the flow, possible values are 'data' and 'exec'"""

        new_alg_mode = FlowAlg.from_str(mode)
        if new_alg_mode is None:
            return False

        self.alg_mode = new_alg_mode
        self._update_running_with_executor()
        self.algorithm_mode_changed.emit(self.algorithm_mode())

        return True


    def _update_running_with_executor(self):
        self.running_with_executor = self.alg_mode in (FlowAlg.DATA_OPT, )

        if self.running_with_executor:
            if self.alg_mode == FlowAlg.DATA_OPT:
                self.executor = self.executor_data_opt
        else:
            self.executor = None


    def flow_changed(self):
        self.executor_data_opt.flow_changed = True


    def data(self) -> dict:
        return {
            'algorithm mode': FlowAlg.str(self.alg_mode),
            'nodes': self.gen_nodes_data(self.nodes),
            'connections': self.gen_conns_data(self.nodes),
            'GID': self.GLOBAL_ID,
        }


    def gen_nodes_data(self, nodes: List[Node]) -> List[dict]:
        """Returns the data dicts of the nodes given"""

        return [n.data() for n in nodes]


    def gen_conns_data(self, nodes: List[Node]) -> List[dict]:
        """Generates the connections data between and relative to the nodes passed"""

        # notice that this is intentionally not part of Connection, because connection data
        # is generated always for a specific set of nodes (like all currently selected ones)
        # and the data dict therefore has the refer to the indices of the nodes in the nodes list

        data = []
        for i in range(len(nodes)):
            n = nodes[i]
            for j in range(len(n.outputs)):
                out = n.outputs[j]
                for c in out.connections:
                    connected_port = c.inp
                    connected_node = connected_port.node

                    # ignore connections connecting to nodes not in the list
                    if connected_node not in nodes:
                        continue

                    data.append({
                        'GID': c.GLOBAL_ID,
                        'parent node index': i,
                        'output port index': j,
                        'connected node': nodes.index(connected_node),
                        'connected input port index': connected_node.inputs.index(connected_port)
                    })

        return data



"""A collection of useful functions used by different components."""

import base64
import pickle
from typing import List


def serialize(data) -> str:
    return base64.b64encode(pickle.dumps(data)).decode('ascii')


def deserialize(data):
    return pickle.loads(base64.b64decode(data))


def node_from_identifier(identifier: str, nodes: List):

    for nc in nodes:
        if nc.identifier == identifier:
            return nc
    else:  # couldn't find a node with this identifier => search for identifier_comp
        for nc in nodes:
            if identifier in nc.identifier_comp:
                return nc
        else:
            raise Exception(
                f'could not find node class with identifier \'{identifier}\'. '
                f'if you changed your node\'s class name, make sure to add the old '
                f'identifier to the identifier_comp list attribute to provide '
                f'backwards compatibility.'
            )
